from __future__ import print_function
import os
import sys
from optparse import OptionParser

from unilint.__version__ import VERSION
from unilint.common import UnilintException
from unilint.common_source_plugin import COMMON_SOURCE
from unilint.unilint_plugin import UnilintPluginInitException
from unilint.custom_format import FLAGS, print_formatted

from unilint.common import run_shell_command

PLUGINS = {}
FORMATTERS = {}

LEVEL_ALL = "all"
LEVEL_STYLE = "style"
LEVEL_WARNING = "warning"
LEVEL_ERROR = "error"

LEVELS = {LEVEL_ALL: 0, LEVEL_STYLE: 10, LEVEL_WARNING: 20, LEVEL_ERROR: 30}


def register_formatter(f_id, formatter_function):
    FORMATTERS[f_id] = formatter_function


def register_plugin(plugin):
    pid = plugin.get_id()
    if pid is None or pid == '':
        raise UnilintException('Invalid plugin id %s'%(pid))
    PLUGINS[pid] = plugin


def extend_maybe(some_container, other_container):
    """extends list or dict with list values with other list or dict unless other is None"""
    if other_container is not None:
        if type(some_container) == dict:
            for key, val in other_container.items():
                if key in some_container:
                    some_container[key].extend(val)
                else:
                    some_container[key] = val
        else:
            some_container.extend(other_container)


def resolve_plugins(selected_plugins_string, deselected_plugins_string, quiet=False):
    """
    Takes in a comma separated String or none, returns a dict str->plugin_instance
    """
    result_plugin_ids = []
    if selected_plugins_string is not None and selected_plugins_string != '':
        selected_plugins_ids = selected_plugins_string.split(',')

        for plugin_id in selected_plugins_ids:
            if plugin_id.strip()=='':
                continue
            if not plugin_id in PLUGINS:
                raise UnilintException('Unknown plugin: %s'%plugin_id)
            result_plugin_ids.append(plugin_id)
    else:
        deselected_plugins_ids = []
        if deselected_plugins_string is not None:
            deselected_plugins_ids = deselected_plugins_string.split(',')

        for plugin_id in PLUGINS:
            if (plugin_id not in deselected_plugins_ids
                and PLUGINS[plugin_id].is_enabled_by_default()):
                result_plugin_ids.append(plugin_id)

    dependencies = []
    for plugin_id in result_plugin_ids:
        deps = PLUGINS[plugin_id].get_depends()
        if deps is not None:
            for dep in deps:
                if not dep in PLUGINS:
                    raise UnilintException('Unknown plugin dependency: %s of %s'%(dep, plugin_id))
                if not dep in result_plugin_ids and not dep in dependencies:
                    dependencies.append(dep)
    for dep_id in dependencies:
        result_plugin_ids.append(dep_id)

    selected_plugins = {}
    for plugin_id in result_plugin_ids:
        try:
            selected_plugins[plugin_id] = PLUGINS[plugin_id](run_shell_command)
        except UnilintPluginInitException as upie:
            if not quiet:
                print("plugin %s failed to activate: %s"%(plugin_id, str(upie)))
    return selected_plugins


def print_issues(options, issues, path):
    if options.formatter is not None:
        FORMATTERS[options.formatter](options, issues, path)
    elif options.format is not None:
        print_formatted(options.format, issues, path)


def order_issues(issues):
    def keyfun(iss):
        return "%s%s"%(iss.path, str(iss.line_number_start).rjust(5, '0'))
    if issues is not None:
        return sorted(issues, key=keyfun)
    return None


def check_path(abspath, basepath, options, subdirs=None, files=None, is_folder=True):
    issues = []
    categorized_resources = {}

    for plugin_id in options.selected_plugins:
        new_categories = options.selected_plugins[plugin_id].categorize_type(options, abspath, subdirs, files)
        extend_maybe(categorized_resources, new_categories)

    for resource, categories in categorized_resources.items():
        if abspath == resource:
            if 'generated' in categories:
                if not options.check_generated:
                    return None
            if 'test' in categories:
                if not options.check_tests:
                    return None
            if 'doc' in categories:
                if not options.check_docs:
                    return None

    for resource, categories in categorized_resources.items():
        if 'hidden' in categories:
            continue
        if 'backup' in categories:
            continue
        if options.debug:
            print("processing: %s as %s"%(resource, categories))
        for plugin_id in options.selected_plugins:
            result = options.selected_plugins[plugin_id].check_resource(options, resource, categories)
            extend_maybe(issues, result)
            if result is not None and not options.ordered and not options.raw:
                # get the output to the user as quickly as possible
                print_issues(options, result, basepath)
    return issues


def run_cmd(path, options, args):
    if options.show_plugins is True:
        print('\n'.join(PLUGINS.keys()))
        return 0

    # if common is not selected, select it. can still be deselected.
    if options.selected_plugins is not None:
        if not COMMON_SOURCE in options.selected_plugins:
            options.selected_plugins += ',%s'%COMMON_SOURCE

    if options.version:
        options.selected_plugins = resolve_plugins(options.selected_plugins,
                                                   options.deselected_plugins,
                                                   quiet=True)

        print("unilint %s"%(VERSION))
        for _, plugin in options.selected_plugins.items():
            meta = plugin.get_meta_information()
            if meta is not None and meta != '':
                print(meta)
        return 0

    options.selected_plugins = resolve_plugins(options.selected_plugins,
                                               options.deselected_plugins)

    if options.debug:
        print('linting from %s'%path)
        print('Activated plugins %s'%options.selected_plugins.keys())

    if len(PLUGINS) == 0:
        raise UnilintException('No plugins registered with unilint')

    if path is None:
        if len(args) == 0:
            raise UnilintException('Unable to use cwd as path')
        else:
            raise UnilintException('Unable to run on path %s'%args[0])

    path = os.path.abspath(path)
    if os.path.isdir(path):
        issues = []

        for (parentdir, subdirs, files) in os.walk(path):
            ignoreddirs = []
            for dirname in subdirs:
                if (dirname in ['build', '.svn', 'CVS', '.hg', '.git']
                    or 'egg-info' in dirname):

                    ignoreddirs.append(dirname)
            for dirname in ignoreddirs:
                subdirs.remove(dirname)
            if len(files) == 0 and len(subdirs) == 0:
                continue
            new_issues = check_path(parentdir, path, options, subdirs, files)
            extend_maybe(issues, new_issues)
            if not options.raw and options.ordered:
                # print here unless we already printed in check_path
                dir_issues = order_issues(new_issues)
                print_issues(options, dir_issues, path)
    else:
        # is file
        issues = check_path(path, os.path.dirname(path), options=options, is_folder=False)
        if not options.raw and options.ordered:
            print_issues(options, issues, path)

    sys.stdout.flush()

    # write to std_err so that report on std_out remains unchanged
    sys.stderr.write("%s issues found\n"%len(issues or []))
    return 0


def get_unilint_parser(usage):
    parser = OptionParser(usage=usage)
    parser.add_option("-v", "--verbose", dest="verbose", default=False,
                      help="verbose output.",
                      action="store_true")
    parser.add_option("--version", dest="version", default=False,
                      help="print version and meta information",
                      action="store_true")
    parser.add_option("--debug", dest="debug", default=False,
                      help="debug output.",
                      action="store_true")
    parser.add_option("-l", "--level", dest="level", default=LEVEL_ERROR,
                      help="A number or one of %s."%(','.join(["%s(=%s)"%(k, v) for k, v in LEVELS.items()])),
                      action="store")
    parser.add_option("-p", "--plugins", dest="show_plugins", default=False,
                      help="List Available plugins",
                      action="store_true")
    parser.add_option("-s", "--select-plugins", dest="selected_plugins", default=None,
                      help="Choose plugins by comma separated id, all others deselected",
                      action="store")
    parser.add_option("-d", "--deselect-plugins", dest="deselected_plugins", default=None,
                      help="Choose plugins by comma separated id, ignored with --select-plugins",
                      action="store")
    parser.add_option("-t", "--tests", dest="check_tests", default=False,
                      help="Also check unit tests",
                      action="store_true")
    parser.add_option("--docs", dest="check_docs", default=False,
                      help="Also check doc files",
                      action="store_true")
    parser.add_option("-g", "--generated", dest="check_generated", default=False,
                      help="Also check generated Files",
                      action="store_true")
    parser.add_option("--format", dest="format", default=None,
                      help="Custom line format using flags %s"%','.join(FLAGS),
                      action="store")
    parser.add_option("-f", "--formatter", dest="formatter", default="brief",
                      help="One of: %s"%','.join(FORMATTERS.keys()),
                      action="store")
    parser.add_option("-o", "--ordered", dest="ordered", default=False,
                      help="Orders issues by file and line",
                      action="store_true")
    parser.add_option("-r", "--raw", dest="raw", default=False,
                      help="Just prints the output of the checker as is",
                      action="store_true")
    return parser


def evaluate_options(options, args):
    path = os.getcwd()
    if len(args) > 0:
        if len(args) > 1:
            raise UnilintException('Too many arguments')

        path = os.path.join(path, args[0])
        if os.path.exists(path):
            path = args[0]
        else:
            path = None

        if options.formatter not in FORMATTERS.keys():
            raise UnilintException("Unknown Formatter: %s"%options.formatter)

    try:
        options.level = int(options.level)
    except ValueError:
        if not options.level in LEVELS:
            raise UnilintException('Unknown level: %s\nAvailable: %s'%(options.level, LEVELS))

        else:
            options.level = LEVELS[options.level]

    return options, args, path


def unilint_main(argv):
    prog = argv[0]
    args = argv[1:]

    parser = get_unilint_parser("usage: %s [OPTIONS] [PATH]"%os.path.basename(prog))
    (options, args) = parser.parse_args(args)

    options, args, path = evaluate_options(options, args)

    return run_cmd(path, options, args)




# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from unilint.python_source_plugin import AbstractPythonPlugin
from unilint.unilint_plugin import UnilintPluginInitException
from unilint.issue import Issue


class PyflakesPlugin(AbstractPythonPlugin):

    def __init__(self, shell_function):
        super(PyflakesPlugin, self).__init__(shell_function)

    @classmethod
    def get_id(cls):
        return 'pyflakes'

    def get_meta_information(self):
        try:
            import pyflakes
        except ImportError as e:
            raise UnilintPluginInitException("ERROR Cannot find pyflakes, install via apt or pip\n%s"%e)
        return "pyflakes: %s"%pyflakes.__version__

    @classmethod
    def is_enabled_by_default(cls):
        # pyflakes returns a subset of pylint, is only useful for quicker checks
        return False

    def check_resource(self, options, path, type_categories):
        if (not 'python-src' in type_categories and
            not 'script' in type_categories):
            return None

        cmd = "pyflakes %s"%path
        if options.verbose:
            print(cmd)
        _, output, _ = self.shell_cmd(cmd,
                                      shell=True,
                                      us_env=True)
        if options.raw:
            print output
            return None
        else:
            return self.parse(output)

    def parse(self, output):
        issues = []
        for line in output.splitlines():
            if line.strip() == '':
                continue
            tokens = line.split(':', 2)
            message = tokens[2].strip()
            severity = "warning"

            issue = Issue(path=tokens[0],
                          message=message,
                          checker_id=self.get_id(),
                          line_number_start=tokens[1],
                          line_number_end=tokens[1],
                          severity=severity)
            issues.append(issue)
        return issues



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from string import Template
from unilint.common import UnilintException

FILENAME = 'f'
RELPATH  = 'F'
MESSAGE  = 'M'
LINE     = 'L'
POSITION = 'P'
CHECKER  = 'C'
TYPE     = 'T'
SEVERITY = 'S'
FLAGSYMBOLS = [FILENAME, RELPATH, MESSAGE, LINE, POSITION, CHECKER, TYPE, SEVERITY]
FLAGS = ['$%s'%flag_temp for flag_temp in FLAGSYMBOLS]


def print_formatted(format_template, issue_list, basepath):
    if format_template is None:
        return
    flag_provided = False

    for flag in FLAGS:
        if flag in format_template:
            flag_provided = True
            break
    if not flag_provided:
        raise UnilintException("No known flags provided : %s"%format_template)

    template = Template(format_template)
    
    for issue in issue_list:
        issuepath = issue.path or basepath
        if basepath != issuepath:
            loc_str = os.path.relpath(issuepath, basepath)
        else:
            loc_str = issuepath
        attr = {FILENAME: os.path.basename(issuepath),
                RELPATH: loc_str,
                MESSAGE: issue.message,
                LINE: issue.line_number_start or '',
                POSITION: issue.line_position or '',
                CHECKER: issue.checker_id,
                TYPE: issue.category or '',
                SEVERITY: issue.severity or ''}
        print(template.substitute(attr))
        



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from unilint.common import is_script

from unilint.unilint_plugin import UnilintPlugin

ROS_SOURCE = 'ros_source'

__pychecker__ = '--unusednames=options,subdirs'


class RosResourcePlugin(UnilintPlugin):
    """Identifies files and folders used by ROS conventions (heuristically)"""

    def __init__(self, shell_function):
        super(RosResourcePlugin, self).__init__(shell_function)

    @classmethod
    def get_id(cls):
        return ROS_SOURCE

    def categorize_type(self, options, path, subdirs, files):
        """
        :returns: list of String
        """
        result = {}
        if files is None:
            if os.path.basename(os.path.dirname(path)) in ['nodes', 'scripts', 'script']:
                if is_script(path):
                    result[path] = ['script']
        else:
            if os.path.basename(path) in ['nodes', 'scripts', 'script']:
                for filename in files:
                    fullname = os.path.join(path, filename)
                    if not filename.endswith('.py') and is_script(fullname):
                        result[fullname] = ['script']
                return result

            if os.path.basename(path) in ['srv', 'msg', 'cfg']:
                is_generated = False
                for filename in files:
                    if filename == '__init__.py':
                        is_generated = True
                if is_generated:
                    result[path] = ['generated']
                    return result
                else:
                    result[path] = [os.path.basename(path)]
            if 'stack.xml' in files:
                result[os.path.join(path)] = ['rosstack']
            if 'manifest.xml' in files:
                result[os.path.join(path)] = ['rospackage']
        return result


class AbstractRosPlugin(UnilintPlugin):
    """Defines a plugin that depends on the categories of RosResourcePlugin"""

    def __init__(self, shell_function):
        super(AbstractRosPlugin, self).__init__(shell_function)

    @classmethod
    def get_depends(cls):
        return ROS_SOURCE



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from unilint.python_source_plugin import AbstractPythonPlugin
from unilint.unilint_plugin import UnilintPluginInitException
from unilint.unilint_main import LEVEL_WARNING, LEVEL_STYLE, LEVELS
from unilint.issue import Issue


class Pep8Plugin(AbstractPythonPlugin):

    def __init__(self, shell_function):
        super(Pep8Plugin, self).__init__(shell_function)

    @classmethod
    def get_id(cls):
        return 'pep8'

    def get_meta_information(self):
        cmd = "pep8 --version"
        value, output, message = self.shell_cmd(cmd,
                                                shell=True,
                                                us_env=True,
                                                ignore_returncodes=[127])
        if value == 0:
            return "pep8: %s"%output
        raise UnilintPluginInitException("ERROR Cannot find pep8, install via apt or pip\n%s"%message)

    def check_resource(self, options, path, type_categories):
        if (not 'python-src' in type_categories and
            not 'script' in type_categories and
            not 'python-module' in type_categories):
            return None

        levels = []

        # pep8 does 'only' concern documentation and style, so we will not use it for level error

        if options.level >= LEVELS[LEVEL_WARNING]:
            if options.verbose:
                print('pep8 not used at this checking level')
            return None

        if options.level < LEVELS[LEVEL_WARNING]:
            levels.append('E')
        if options.level < LEVELS[LEVEL_STYLE]:
            levels.append('W')

        cmd = "pep8 %s -r --select=%s"%(path, ','.join(levels))

        if options.verbose:
            print(cmd)

        _, output, _ = self.shell_cmd(cmd,
                                      shell=True,
                                      us_env=True)

        if options.raw:
            print output
            return None
        else:
            return self.parse(output)

    def parse(self, output):
        issues = []
        for line in output.splitlines():
            if line.strip() == '':
                continue
            tokens = line.split(':', 3)
            (category, message) = tokens[3].strip().split(' ', 1)
            severity=None
            if category.startswith('E'):
                severity = "warning"
            if category.startswith('W'):
                severity = "style"

            issue = Issue(path=tokens[0],
                          message=message,
                          checker_id=self.get_id(),
                          line_number_start=tokens[1],
                          line_number_end=tokens[1],
                          line_position=tokens[2],
                          severity=severity,
                          category=category)
            issues.append(issue)
        return issues



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os

from unilint.unilint_main import register_formatter
from unilint.custom_format import print_formatted

# pylint: disable-msg=W0613


def register_standard_formatters():
    def print_brief(unused_options, issue_list, basepath):
        if issue_list is None:
            return
        for issue in issue_list:
            issuepath = issue.path or basepath
            if basepath != issuepath:
                loc_str = os.path.relpath(issuepath, basepath)
            else:
                loc_str = issuepath
            location="%s :%s:%s:"%(loc_str,
                                   str(issue.line_number_start or '').ljust(3),
                                   str(issue.line_position or '').ljust(3))
            location = location.ljust(20)
            print("%s   %s"%(location, issue.message))

    def print_full(unused_options, issue_list, basepath):
        print_formatted('[$C\t$S\t$T]$F:$L:$P:\t$M', issue_list, basepath)

    def print_short(unused_options, issue_list, basepath):
        print_formatted('$f:$L: $M', issue_list, basepath)

    register_formatter('brief', print_brief)
    register_formatter('full', print_full)
    register_formatter('short', print_short)



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""module unilint"""

import logging

from unilint.unilint_main import register_plugin

from unilint.common_source_plugin import CommonSourcePlugin
from unilint.python_source_plugin import PythonSourcePlugin
from unilint.cpp_source_plugin import CppSourcePlugin

from unilint.pep8_plugin import Pep8Plugin
from unilint.pyflakes_plugin import PyflakesPlugin
from unilint.pylint_plugin import PylintPlugin
from unilint.pychecker_plugin import PycheckerPlugin

from unilint.cppcheck_plugin import CppcheckPlugin

from unilint.standard_formatters import register_standard_formatters

register_standard_formatters()

register_plugin(CommonSourcePlugin)
register_plugin(PythonSourcePlugin)
register_plugin(CppSourcePlugin)

register_plugin(Pep8Plugin)
register_plugin(PyflakesPlugin)
register_plugin(PycheckerPlugin)
register_plugin(PylintPlugin)
register_plugin(CppcheckPlugin)


def setup_logger():
    logger = logging.getLogger('unilint')
    logger.setLevel(logging.WARN)
    message_handler = logging.StreamHandler()
    message_handler.setLevel(logging.WARN)

    # create formatter
    formatter = logging.Formatter('%(levelname)s [%(name)s] %(message)s[/%(name)s]')
    # add formatter to message_handler
    message_handler.setFormatter(formatter)

    # add message_handler to logger
    logger.addHandler(message_handler)

setup_logger()



# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Thibault Kruse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import re
import logging
import subprocess
import copy


class UnilintException(Exception):
    pass


def is_binary(filename):
    """Return true if the given filename is binary."""
    fin = open(filename, 'rb')
    try:
        chunksize = 1024
        while 1:
            chunk = fin.read(chunksize)
            if '\0' in chunk:  # found null byte
                return True
            if len(chunk) < chunksize:
                break  # done
    finally:
        fin.close()
    return False


def is_script(path):
    return (re.match('^#|.*~$|^_|^\.', os.path.basename(path)) is None and
            not is_binary(path))


def run_shell_command(cmd,
                      cwd=None,
                      shell=False,
                      us_env=True,
                      show_stdout=False,
                      verbose=False,
                      ignore_returncodes=None,
                      capture_std_err=False):
    """
    executes a command and hides the stdout output, loggs
    stderr output when command result is not zero. Make sure to sanitize arguments in the command.

    :param cmd: A string to execute.
    :param shell: Whether to use os shell, this is DANGEROUS, as vulnerable to shell injection.
    :returns: ( returncode, stdout, stderr)
    :raises: OSError
    """
    try:
        env = copy.copy(os.environ)
        if us_env:
            env["LANG"] = "en_US.UTF-8"
        proc = subprocess.Popen(cmd,
                             shell=shell,
                             cwd=cwd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             env=env)
        # when we read output in while loop, it will not be returned in communicate()
        stderr_buf = []
        stdout_buf = []
        if verbose or show_stdout:
            # listen to stdout and print
            while True:
                line = proc.stdout.readline()
                if line != '':
                    if verbose:
                        print(line),
                        stdout_buf.append(line)
                line2 = proc.stderr.readline()
                if line2 != '':
                    if verbose:
                        print(line2),
                        stderr_buf.append(line2)
                if ((not line and not line2) or proc.returncode is not None):
                    break
        (stdout, stderr) = proc.communicate()
        stdout_buf.append(stdout)
        stdout = "\n".join(stdout_buf)
        stderr_buf.append(stderr)
        stderr = "\n".join(stderr_buf)
        if capture_std_err is True:
            stdout = "\n".join(stderr_buf)

        message = None
        if (proc.returncode != 0 and
            (ignore_returncodes is None or
             not proc.returncode in ignore_returncodes) and
            stderr is not None and stderr != ''):
            logger = logging.getLogger('unilint')
            message = "Command failed: '%s'"%(cmd)
            if cwd is not None:
                message += "\n run at: '%s'"%(cwd)
            message += "\n errcode: %s:\n%s"%(proc.returncode, stderr)
            logger.warn(message)
        result = stdout
        if result is not None:
            result = result.rstrip()
        return (proc.returncode, result, message)
    except OSError as ose:
        logger = logging.getLogger('unilint')
        message = "Command failed with OSError. '%s' <%s, %s>:\n%s"%(cmd, shell, cwd, ose)
        logger.error(message)
        raise
